"""implementations of simple readline edit operations

just the ones that fit the model of transforming the current line
and the cursor location
based on http://www.bigsmoke.us/readline/shortcuts"""

import inspect

from ..lazyre import LazyReCompile
from ..line import cursor_on_closing_char_pair

INDENT = 4

# TODO Allow user config of keybindings for these actions
getargspec = lambda func: inspect.signature(func).parameters


class AbstractEdits:
    default_kwargs = {
        "line": "hello world",
        "cursor_offset": 5,
        "cut_buffer": "there",
    }

    def __init__(self, simple_edits=None, cut_buffer_edits=None):
        self.simple_edits = {} if simple_edits is None else simple_edits
        self.cut_buffer_edits = (
            {} if cut_buffer_edits is None else cut_buffer_edits
        )
        self.awaiting_config = {}

    def add(self, key, func, overwrite=False):
        if key in self:
            if overwrite:
                del self[key]
            else:
                raise ValueError(f"key {key!r} already has a mapping")
        params = getargspec(func)
        args = {k: v for k, v in self.default_kwargs.items() if k in params}
        r = func(**args)
        if len(r) == 2:
            if hasattr(func, "kills"):
                raise ValueError(
                    "function %r returns two values, but has a "
                    "kills attribute" % (func,)
                )
            self.simple_edits[key] = func
        elif len(r) == 3:
            if not hasattr(func, "kills"):
                raise ValueError(
                    "function %r returns three values, but has "
                    "no kills attribute" % (func,)
                )
            self.cut_buffer_edits[key] = func
        else:
            raise ValueError(f"return type of function {func!r} not recognized")

    def add_config_attr(self, config_attr, func):
        if config_attr in self.awaiting_config:
            raise ValueError(
                f"config attribute {config_attr!r} already has a mapping"
            )
        self.awaiting_config[config_attr] = func

    def call(self, key, **kwargs):
        func = self[key]
        params = getargspec(func)
        args = {k: v for k, v in kwargs.items() if k in params}
        return func(**args)

    def __contains__(self, key):
        return key in self.simple_edits or key in self.cut_buffer_edits

    def __getitem__(self, key):
        if key in self.simple_edits:
            return self.simple_edits[key]
        if key in self.cut_buffer_edits:
            return self.cut_buffer_edits[key]
        raise KeyError(f"key {key!r} not mapped")

    def __delitem__(self, key):
        if key in self.simple_edits:
            del self.simple_edits[key]
        elif key in self.cut_buffer_edits:
            del self.cut_buffer_edits[key]
        else:
            raise KeyError(f"key {key!r} not mapped")


class UnconfiguredEdits(AbstractEdits):
    """Maps key to edit functions, and bins them by what parameters they take.

    Only functions with specific signatures can be added:
        * func(**kwargs) -> cursor_offset, line
        * func(**kwargs) -> cursor_offset, line, cut_buffer
        where kwargs are in among the keys of Edits.default_kwargs
    These functions will be run to determine their return type, so no side
    effects!

    More concrete Edits instances can be created by applying a config with
    Edits.mapping_with_config() - this creates a new Edits instance
    that uses a config file to assign config_attr bindings.

    Keys can't be added twice, config attributes can't be added twice.
    """

    def mapping_with_config(self, config, key_dispatch):
        """Creates a new mapping object by applying a config object"""
        return ConfiguredEdits(
            self.simple_edits,
            self.cut_buffer_edits,
            self.awaiting_config,
            config,
            key_dispatch,
        )

    def on(self, key=None, config=None):
        if not ((key is None) ^ (config is None)):
            raise ValueError("Must use exactly one of key, config")
        if key is not None:

            def add_to_keybinds(func):
                self.add(key, func)
                return func

            return add_to_keybinds
        else:

            def add_to_config(func):
                self.add_config_attr(config, func)
                return func

            return add_to_config


class ConfiguredEdits(AbstractEdits):
    def __init__(
        self,
        simple_edits,
        cut_buffer_edits,
        awaiting_config,
        config,
        key_dispatch,
    ):
        super().__init__(dict(simple_edits), dict(cut_buffer_edits))
        for attr, func in awaiting_config.items():
            for key in key_dispatch[getattr(config, attr)]:
                super().add(key, func, overwrite=True)

    def add_config_attr(self, config_attr, func):
        raise NotImplementedError("Config already set on this mapping")

    def add(self, key, func, overwrite=False):
        raise NotImplementedError("Config already set on this mapping")


edit_keys = UnconfiguredEdits()

# Because the edits.on decorator runs the functions, functions which depend
# on other functions must be declared after their dependencies


def kills_behind(func):
    func.kills = "behind"
    return func


def kills_ahead(func):
    func.kills = "ahead"
    return func


@edit_keys.on(config="left_key")
@edit_keys.on("<LEFT>")
def left_arrow(cursor_offset, line):
    return max(0, cursor_offset - 1), line


@edit_keys.on(config="right_key")
@edit_keys.on("<RIGHT>")
def right_arrow(cursor_offset, line):
    return min(len(line), cursor_offset + 1), line


@edit_keys.on(config="beginning_of_line_key")
@edit_keys.on("<HOME>")
def beginning_of_line(cursor_offset, line):
    return 0, line


@edit_keys.on(config="end_of_line_key")
@edit_keys.on("<END>")
def end_of_line(cursor_offset, line):
    return len(line), line


forward_word_re = LazyReCompile(r"\S\s")


@edit_keys.on("<Esc+f>")
@edit_keys.on("<Ctrl-RIGHT>")
@edit_keys.on("<Esc+RIGHT>")
def forward_word(cursor_offset, line):
    match = forward_word_re.search(line[cursor_offset:] + " ")
    delta = match.end() - 1 if match else 0
    return (cursor_offset + delta, line)


def last_word_pos(string):
    """returns the start index of the last word of given string"""
    match = forward_word_re.search(string[::-1])
    index = match and len(string) - match.end() + 1
    return index or 0


@edit_keys.on("<Esc+b>")
@edit_keys.on("<Ctrl-LEFT>")
@edit_keys.on("<Esc+LEFT>")
def back_word(cursor_offset, line):
    return (last_word_pos(line[:cursor_offset]), line)


@edit_keys.on("<DELETE>")
def delete(cursor_offset, line):
    return (cursor_offset, line[:cursor_offset] + line[cursor_offset + 1 :])


@edit_keys.on("<BACKSPACE>")
@edit_keys.on(config="backspace_key")
def backspace(cursor_offset, line):
    if cursor_offset == 0:
        return cursor_offset, line
    if not line[:cursor_offset].strip():  # if just whitespace left of cursor
        # front_white = len(line[:cursor_offset]) - \
        #     len(line[:cursor_offset].lstrip())
        to_delete = ((cursor_offset - 1) % INDENT) + 1
        return (
            cursor_offset - to_delete,
            line[: cursor_offset - to_delete] + line[cursor_offset:],
        )
    # removes opening bracket along with closing bracket
    # if there is nothing between them
    # TODO: could not get config value here, works even without -B option
    on_closing_char, pair_close = cursor_on_closing_char_pair(
        cursor_offset, line
    )
    if on_closing_char and pair_close:
        return (
            cursor_offset - 1,
            line[: cursor_offset - 1] + line[cursor_offset + 1 :],
        )

    return (cursor_offset - 1, line[: cursor_offset - 1] + line[cursor_offset:])


@edit_keys.on(config="clear_line_key")
def delete_from_cursor_back(cursor_offset, line):
    return 0, line[cursor_offset:]


delete_rest_of_word_re = LazyReCompile(r"\w\b")


@edit_keys.on("<Esc+d>")  # option-d
@kills_ahead
def delete_rest_of_word(cursor_offset, line):
    m = delete_rest_of_word_re.search(line[cursor_offset:])
    if not m:
        return cursor_offset, line, ""
    return (
        cursor_offset,
        line[:cursor_offset] + line[m.start() + cursor_offset + 1 :],
        line[cursor_offset : m.start() + cursor_offset + 1],
    )


delete_word_to_cursor_re = LazyReCompile(r"\s\S")


@edit_keys.on(config="clear_word_key")
@kills_behind
def delete_word_to_cursor(cursor_offset, line):
    start = 0
    for match in delete_word_to_cursor_re.finditer(line[:cursor_offset]):
        start = match.start() + 1
    return (
        start,
        line[:start] + line[cursor_offset:],
        line[start:cursor_offset],
    )


@edit_keys.on("<Esc+y>")
def yank_prev_prev_killed_text(cursor_offset, line, cut_buffer):
    # TODO not implemented - just prev
    return (
        cursor_offset + len(cut_buffer),
        line[:cursor_offset] + cut_buffer + line[cursor_offset:],
    )


@edit_keys.on(config="yank_from_buffer_key")
def yank_prev_killed_text(cursor_offset, line, cut_buffer):
    return (
        cursor_offset + len(cut_buffer),
        line[:cursor_offset] + cut_buffer + line[cursor_offset:],
    )


@edit_keys.on(config="transpose_chars_key")
def transpose_character_before_cursor(cursor_offset, line):
    if cursor_offset < 2:
        return cursor_offset, line
    if cursor_offset == len(line):
        return cursor_offset, line[:-2] + line[-1] + line[-2]
    return (
        min(len(line), cursor_offset + 1),
        line[: cursor_offset - 1]
        + (line[cursor_offset] if len(line) > cursor_offset else "")
        + line[cursor_offset - 1]
        + line[cursor_offset + 1 :],
    )


@edit_keys.on("<Esc+t>")
def transpose_word_before_cursor(cursor_offset, line):
    return cursor_offset, line  # TODO Not implemented


# TODO undo all changes to line: meta-r

# bonus functions (not part of readline)


@edit_keys.on("<Esc+u>")
def uppercase_next_word(cursor_offset, line):
    return cursor_offset, line  # TODO Not implemented


@edit_keys.on(config="cut_to_buffer_key")
@kills_ahead
def delete_from_cursor_forward(cursor_offset, line):
    return cursor_offset, line[:cursor_offset], line[cursor_offset:]


@edit_keys.on("<Esc+c>")
def titlecase_next_word(cursor_offset, line):
    return cursor_offset, line  # TODO Not implemented


delete_word_from_cursor_back_re = LazyReCompile(r"^|\b\w")


@edit_keys.on("<Esc+BACKSPACE>")
@edit_keys.on("<Meta-BACKSPACE>")
@kills_behind
def delete_word_from_cursor_back(cursor_offset, line):
    """Whatever my option-delete does in bash on my mac"""
    if not line:
        return cursor_offset, line, ""
    start = None
    for match in delete_word_from_cursor_back_re.finditer(line):
        if match.start() < cursor_offset:
            start = match.start()
    if start is not None:
        return (
            start,
            line[:start] + line[cursor_offset:],
            line[start:cursor_offset],
        )
    else:
        return cursor_offset, line, ""



import code
import os
import sys
import tempfile
import io
from typing import cast
import unittest

from contextlib import contextmanager
from functools import partial
from unittest import mock

from bpython.curtsiesfrontend import repl as curtsiesrepl
from bpython.curtsiesfrontend import interpreter
from bpython.curtsiesfrontend import events as bpythonevents
from bpython.curtsiesfrontend.repl import LineType
from bpython import autocomplete
from bpython import config
from bpython import args
from bpython.test import (
    FixLanguageTestCase as TestCase,
    MagicIterMock,
    TEST_CONFIG,
)

from curtsies import events
from curtsies.window import CursorAwareWindow
from importlib import invalidate_caches


def setup_config(conf):
    config_struct = config.Config(TEST_CONFIG)
    for key, value in conf.items():
        if not hasattr(config_struct, key):
            raise ValueError(f"{key!r} is not a valid config attribute")
        setattr(config_struct, key, value)
    return config_struct


class TestCurtsiesRepl(TestCase):
    def setUp(self):
        self.repl = create_repl()

    def cfwp(self, source):
        return interpreter.code_finished_will_parse(
            source, self.repl.interp.compile
        )

    def test_code_finished_will_parse(self):
        self.repl.buffer = ["1 + 1"]
        self.assertTrue(self.cfwp("\n".join(self.repl.buffer)), (True, True))
        self.repl.buffer = ["def foo(x):"]
        self.assertTrue(self.cfwp("\n".join(self.repl.buffer)), (False, True))
        self.repl.buffer = ["def foo(x)"]
        self.assertTrue(self.cfwp("\n".join(self.repl.buffer)), (True, False))
        self.repl.buffer = ["def foo(x):", "return 1"]
        self.assertTrue(self.cfwp("\n".join(self.repl.buffer)), (True, False))
        self.repl.buffer = ["def foo(x):", "    return 1"]
        self.assertTrue(self.cfwp("\n".join(self.repl.buffer)), (True, True))
        self.repl.buffer = ["def foo(x):", "    return 1", ""]
        self.assertTrue(self.cfwp("\n".join(self.repl.buffer)), (True, True))

    def test_external_communication(self):
        self.repl.send_current_block_to_external_editor()
        self.repl.send_session_to_external_editor()

    @unittest.skipUnless(
        all(map(config.can_encode, "å∂ßƒ")), "Charset can not encode characters"
    )
    def test_external_communication_encoding(self):
        with captured_output():
            self.repl.display_lines.append('>>> "åß∂ƒ"')
            self.repl.history.append('"åß∂ƒ"')
            self.repl.all_logical_lines.append(('"åß∂ƒ"', LineType.INPUT))
            self.repl.send_session_to_external_editor()

    def test_get_last_word(self):
        self.repl.rl_history.entries = ["1", "2 3", "4 5 6"]
        self.repl._set_current_line("abcde")
        self.repl.get_last_word()
        self.assertEqual(self.repl.current_line, "abcde6")
        self.repl.get_last_word()
        self.assertEqual(self.repl.current_line, "abcde3")

    def test_last_word(self):
        self.assertEqual(curtsiesrepl._last_word(""), "")
        self.assertEqual(curtsiesrepl._last_word(" "), "")
        self.assertEqual(curtsiesrepl._last_word("a"), "a")
        self.assertEqual(curtsiesrepl._last_word("a b"), "b")

    @unittest.skip("this is the behavior of bash - not currently implemented")
    def test_get_last_word_with_prev_line(self):
        self.repl.rl_history.entries = ["1", "2 3", "4 5 6"]
        self.repl._set_current_line("abcde")
        self.repl.up_one_line()
        self.assertEqual(self.repl.current_line, "4 5 6")
        self.repl.get_last_word()
        self.assertEqual(self.repl.current_line, "4 5 63")
        self.repl.get_last_word()
        self.assertEqual(self.repl.current_line, "4 5 64")
        self.repl.up_one_line()
        self.assertEqual(self.repl.current_line, "2 3")


def mock_next(obj, return_value):
    obj.__next__.return_value = return_value


class TestCurtsiesReplTab(TestCase):
    def setUp(self):
        self.repl = create_repl()
        self.repl.matches_iter = MagicIterMock()

        def add_matches(*args, **kwargs):
            self.repl.matches_iter.matches = ["aaa", "aab", "aac"]

        self.repl.complete = mock.Mock(
            side_effect=add_matches, return_value=True
        )

    def test_tab_with_no_matches_triggers_completion(self):
        self.repl._current_line = " asdf"
        self.repl._cursor_offset = 5
        self.repl.matches_iter.matches = []
        self.repl.matches_iter.is_cseq.return_value = False
        self.repl.matches_iter.cur_line.return_value = (None, None)
        self.repl.on_tab()
        self.repl.complete.assert_called_once_with(tab=True)

    def test_tab_after_indentation_adds_space(self):
        self.repl._current_line = "    "
        self.repl._cursor_offset = 4
        self.repl.on_tab()
        self.assertEqual(self.repl._current_line, "        ")
        self.assertEqual(self.repl._cursor_offset, 8)

    def test_tab_at_beginning_of_line_adds_space(self):
        self.repl._current_line = ""
        self.repl._cursor_offset = 0
        self.repl.on_tab()
        self.assertEqual(self.repl._current_line, "    ")
        self.assertEqual(self.repl._cursor_offset, 4)

    def test_tab_with_no_matches_selects_first(self):
        self.repl._current_line = " aa"
        self.repl._cursor_offset = 3
        self.repl.matches_iter.matches = []
        self.repl.matches_iter.is_cseq.return_value = False

        mock_next(self.repl.matches_iter, None)
        self.repl.matches_iter.cur_line.return_value = (None, None)
        self.repl.on_tab()
        self.repl.complete.assert_called_once_with(tab=True)
        self.repl.matches_iter.cur_line.assert_called_once_with()

    def test_tab_with_matches_selects_next_match(self):
        self.repl._current_line = " aa"
        self.repl._cursor_offset = 3
        self.repl.complete()
        self.repl.matches_iter.is_cseq.return_value = False
        mock_next(self.repl.matches_iter, None)
        self.repl.matches_iter.cur_line.return_value = (None, None)
        self.repl.on_tab()
        self.repl.matches_iter.cur_line.assert_called_once_with()

    def test_tab_completes_common_sequence(self):
        self.repl._current_line = " a"
        self.repl._cursor_offset = 2
        self.repl.matches_iter.matches = ["aaa", "aab", "aac"]
        self.repl.matches_iter.is_cseq.return_value = True
        self.repl.matches_iter.substitute_cseq.return_value = (None, None)
        self.repl.on_tab()
        self.repl.matches_iter.substitute_cseq.assert_called_once_with()


class TestCurtsiesReplFilenameCompletion(TestCase):
    def setUp(self):
        self.repl = create_repl()

    def test_list_win_visible_match_selected_on_tab_multiple_options(self):
        self.repl._current_line = " './'"
        self.repl._cursor_offset = 2
        with mock.patch("bpython.autocomplete.get_completer") as m:
            m.return_value = (
                ["./abc", "./abcd", "./bcd"],
                autocomplete.FilenameCompletion(),
            )
            self.repl.update_completion()
            self.assertEqual(self.repl.list_win_visible, False)
            self.repl.on_tab()
        self.assertEqual(self.repl.current_match, "./abc")
        self.assertEqual(self.repl.list_win_visible, True)

    def test_list_win_not_visible_and_cseq_if_cseq(self):
        self.repl._current_line = " './a'"
        self.repl._cursor_offset = 5
        with mock.patch("bpython.autocomplete.get_completer") as m:
            m.return_value = (
                ["./abcd", "./abce"],
                autocomplete.FilenameCompletion(),
            )
            self.repl.update_completion()
            self.assertEqual(self.repl.list_win_visible, False)
        self.repl.on_tab()
        self.assertEqual(self.repl._current_line, " './abc'")
        self.assertEqual(self.repl.current_match, None)
        self.assertEqual(self.repl.list_win_visible, False)

    def test_list_win_not_visible_and_match_selected_if_one_option(self):
        self.repl._current_line = " './a'"
        self.repl._cursor_offset = 5
        with mock.patch("bpython.autocomplete.get_completer") as m:
            m.return_value = (["./abcd"], autocomplete.FilenameCompletion())
            self.repl.update_completion()
            self.assertEqual(self.repl.list_win_visible, False)
        self.repl.on_tab()
        self.assertEqual(self.repl._current_line, " './abcd'")
        self.assertEqual(self.repl.current_match, None)
        self.assertEqual(self.repl.list_win_visible, False)


# from http://stackoverflow.com/a/17981937/398212 - thanks @rkennedy
@contextmanager
def captured_output():
    new_out, new_err = io.StringIO(), io.StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


def create_repl(**kwargs):
    config = setup_config({"editor": "true"})
    repl = curtsiesrepl.BaseRepl(
        config, cast(CursorAwareWindow, None), **kwargs
    )
    os.environ["PAGER"] = "true"
    os.environ.pop("PYTHONSTARTUP", None)
    repl.width = 50
    repl.height = 20
    return repl


class TestFutureImports(TestCase):
    def test_repl(self):
        repl = create_repl()
        with captured_output() as (out, err):
            repl.push("1 / 2")
        self.assertEqual(out.getvalue(), "0.5\n")

    def test_interactive(self):
        interp = code.InteractiveInterpreter(locals={})
        with captured_output() as (out, err):
            with tempfile.NamedTemporaryFile(mode="w", suffix=".py") as f:
                f.write("print(1/2)\n")
                f.flush()
                args.exec_code(interp, [f.name])

            repl = create_repl(interp=interp)
            repl.push("1 / 2")

        self.assertEqual(out.getvalue(), "0.5\n0.5\n")


class TestStdOutErr(TestCase):
    def setUp(self):
        self.repl = create_repl()

    def test_newline(self):
        self.repl.send_to_stdouterr("\n\n")
        self.assertEqual(self.repl.display_lines[-2], "")
        self.assertEqual(self.repl.display_lines[-1], "")
        self.assertEqual(self.repl.current_stdouterr_line, "")

    def test_leading_newline(self):
        self.repl.send_to_stdouterr("\nfoo\n")
        self.assertEqual(self.repl.display_lines[-2], "")
        self.assertEqual(self.repl.display_lines[-1], "foo")
        self.assertEqual(self.repl.current_stdouterr_line, "")

    def test_no_trailing_newline(self):
        self.repl.send_to_stdouterr("foo")
        self.assertEqual(self.repl.current_stdouterr_line, "foo")

    def test_print_without_newline_then_print_with_leading_newline(self):
        self.repl.send_to_stdouterr("foo")
        self.repl.send_to_stdouterr("\nbar\n")
        self.assertEqual(self.repl.display_lines[-2], "foo")
        self.assertEqual(self.repl.display_lines[-1], "bar")
        self.assertEqual(self.repl.current_stdouterr_line, "")


class TestPredictedIndent(TestCase):
    def setUp(self):
        self.repl = create_repl()

    def test_simple(self):
        self.assertEqual(self.repl.predicted_indent(""), 0)
        self.assertEqual(self.repl.predicted_indent("class Foo:"), 4)
        self.assertEqual(self.repl.predicted_indent("class Foo: pass"), 0)
        self.assertEqual(self.repl.predicted_indent("def asdf():"), 4)
        self.assertEqual(self.repl.predicted_indent("def asdf(): return 7"), 0)

    @unittest.skip("This would be interesting")
    def test_complex(self):
        self.assertEqual(self.repl.predicted_indent("[a, "), 1)
        self.assertEqual(self.repl.predicted_indent("reduce(asdfasdf, "), 7)


class TestCurtsiesReevaluate(TestCase):
    def setUp(self):
        self.repl = create_repl()

    def test_variable_is_cleared(self):
        self.repl._current_line = "b = 10"
        self.repl.on_enter()
        self.assertIn("b", self.repl.interp.locals)
        self.repl.undo()
        self.assertNotIn("b", self.repl.interp.locals)


class TestCurtsiesReevaluateWithImport(TestCase):
    def setUp(self):
        self.repl = create_repl()
        self.open = partial(io.open, mode="wt", encoding="utf-8")
        self.dont_write_bytecode = sys.dont_write_bytecode
        sys.dont_write_bytecode = True
        self.sys_path = sys.path
        sys.path = self.sys_path[:]

        # Because these tests create Python source files at runtime,
        # it's possible in Python >=3.3 for the importlib.machinery.FileFinder
        # for a directory to have an outdated cache when
        # * a module in that directory is imported,
        # * then a new module is created in that directory,
        # * then that new module is imported.
        # Automatic cache invalidation is based on the second-resolution mtime
        # of the directory, so we need to manually call invalidate_caches().
        #
        # see https://docs.python.org/3/library/importlib.html
        # sections #importlib.machinery.FileFinder and
        # #importlib.invalidate_caches
        invalidate_caches()

    def tearDown(self):
        sys.dont_write_bytecode = self.dont_write_bytecode
        sys.path = self.sys_path

    def push(self, line):
        self.repl._current_line = line
        self.repl.on_enter()

    def head(self, path):
        self.push("import sys")
        self.push('sys.path.append("%s")' % (path))

    @staticmethod
    @contextmanager
    def tempfile():
        with tempfile.NamedTemporaryFile(suffix=".py") as temp:
            path, name = os.path.split(temp.name)
            yield temp.name, path, name.replace(".py", "")

    def test_module_content_changed(self):
        with self.tempfile() as (fullpath, path, modname):
            print(modname)
            with self.open(fullpath) as f:
                f.write("a = 0\n")
            self.head(path)
            self.push("import %s" % (modname))
            self.push("a = %s.a" % (modname))
            self.assertIn("a", self.repl.interp.locals)
            self.assertEqual(self.repl.interp.locals["a"], 0)
            with self.open(fullpath) as f:
                f.write("a = 1\n")
            self.repl.clear_modules_and_reevaluate()
            self.assertIn("a", self.repl.interp.locals)
            self.assertEqual(self.repl.interp.locals["a"], 1)

    def test_import_module_with_rewind(self):
        with self.tempfile() as (fullpath, path, modname):
            print(modname)
            with self.open(fullpath) as f:
                f.write("a = 0\n")
            self.head(path)
            self.push("import %s" % (modname))
            self.assertIn(modname, self.repl.interp.locals)
            self.repl.undo()
            self.assertNotIn(modname, self.repl.interp.locals)
            self.repl.clear_modules_and_reevaluate()
            self.assertNotIn(modname, self.repl.interp.locals)
            self.push("import %s" % (modname))
            self.push("a = %s.a" % (modname))
            self.assertIn("a", self.repl.interp.locals)
            self.assertEqual(self.repl.interp.locals["a"], 0)
            with self.open(fullpath) as f:
                f.write("a = 1\n")
            self.repl.clear_modules_and_reevaluate()
            self.assertIn("a", self.repl.interp.locals)
            self.assertEqual(self.repl.interp.locals["a"], 1)


class TestCurtsiesPagerText(TestCase):
    def setUp(self):
        self.repl = create_repl()
        self.repl.pager = self.assert_pager_gets_unicode

    def assert_pager_gets_unicode(self, text):
        self.assertIsInstance(text, str)

    def test_help(self):
        self.repl.pager(self.repl.help_text())

    @unittest.skipUnless(
        all(map(config.can_encode, "å∂ßƒ")), "Charset can not encode characters"
    )
    def test_show_source_not_formatted(self):
        self.repl.config.highlight_show_source = False
        self.repl.get_source_of_current_name = lambda: "source code å∂ßƒåß∂ƒ"
        self.repl.show_source()

    @unittest.skipUnless(
        all(map(config.can_encode, "å∂ßƒ")), "Charset can not encode characters"
    )
    def test_show_source_formatted(self):
        self.repl.config.highlight_show_source = True
        self.repl.get_source_of_current_name = lambda: "source code å∂ßƒåß∂ƒ"
        self.repl.show_source()


class TestCurtsiesStartup(TestCase):
    def setUp(self):
        self.repl = create_repl()

    def write_startup_file(self, fname, encoding):
        with open(fname, mode="w", encoding=encoding) as f:
            f.write("# coding: ")
            f.write(encoding)
            f.write("\n")
            f.write('a = "äöü"\n')

    def test_startup_event_utf8(self):
        with tempfile.NamedTemporaryFile() as temp:
            self.write_startup_file(temp.name, "utf-8")
            with mock.patch.dict("os.environ", {"PYTHONSTARTUP": temp.name}):
                self.repl.process_event(bpythonevents.RunStartupFileEvent())
        self.assertIn("a", self.repl.interp.locals)

    def test_startup_event_latin1(self):
        with tempfile.NamedTemporaryFile() as temp:
            self.write_startup_file(temp.name, "latin-1")
            with mock.patch.dict("os.environ", {"PYTHONSTARTUP": temp.name}):
                self.repl.process_event(bpythonevents.RunStartupFileEvent())
        self.assertIn("a", self.repl.interp.locals)


class TestCurtsiesPasteEvents(TestCase):
    def setUp(self):
        self.repl = create_repl()

    def test_control_events_in_small_paste(self):
        self.assertGreaterEqual(
            curtsiesrepl.MAX_EVENTS_POSSIBLY_NOT_PASTE,
            6,
            "test assumes UI lag could cause 6 events",
        )
        p = events.PasteEvent()
        p.events = ["a", "b", "c", "d", "<Ctrl-a>", "e"]
        self.repl.process_event(p)
        self.assertEqual(self.repl.current_line, "eabcd")

    def test_control_events_in_large_paste(self):
        """Large paste events should ignore control characters"""
        p = events.PasteEvent()
        p.events = ["a", "<Ctrl-a>"] + [
            "e"
        ] * curtsiesrepl.MAX_EVENTS_POSSIBLY_NOT_PASTE
        self.repl.process_event(p)
        self.assertEqual(
            self.repl.current_line,
            "a" + "e" * curtsiesrepl.MAX_EVENTS_POSSIBLY_NOT_PASTE,
        )


if __name__ == "__main__":
    unittest.main()
